#include "calfile.h"
#include <zlib.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>

/* string versions of the calerr enum */
const char *cal_err_strings[] = {
  "no error",
  "checksum failed",
  "file open failed",
  "invalid call order",
  "unsupported call",
  "write attempt to read-only file",
  "read attempt on write-only file",
  "file IO error",
  "error invalid argument",
  "file has the wrong magic",
  "file version unsupported",
  "calfile struct corrupted",
  "tried to stat a symlink -- internal error",
  "file is an unexpected size",
  "string size too large",
};

/*
 *   UTIL FUNCS
 */

#define FLOATTOLONG(a) (*(unsigned long*)&a)
#define LONGTOFLOAT(a) (*(float*)&a)

inline int readu32(int fd, uint32_t *val) {
  int ret;
  ret = read(fd, val, sizeof(uint32_t)) == sizeof(uint32_t);
  *val = ntohl(*val);
  return ret;
}

inline int readu16(int fd, uint16_t *val) {
  int ret;
  ret = read(fd, val, sizeof(uint16_t)) == sizeof(uint16_t);
  *val = ntohs(*val);
  return ret;
}

inline int writeu32(int fd, uint32_t val) {
  int ret;
  val = htonl(val);
  ret = write(fd, &val, sizeof(uint32_t)) == sizeof(uint32_t);
  return ret;
}

inline int writeu16(int fd, uint16_t val) {
  int ret;
  val = htons(val);
  ret = write(fd, &val, sizeof(uint16_t)) == sizeof(uint16_t);
  return ret;
}

#define CRC_BUF_SIZE 512
static int calc_crc(struct calfile *cal, uint32_t *crc) {
  char buf[CRC_BUF_SIZE];
  ssize_t count;

  if (lseek(cal->fd, 8, SEEK_SET) != 8) {
    cal->err = CAL_ERR_IO;
    return 0;
  }

  *crc = crc32(0, Z_NULL, 0);
  while ((count = read(cal->fd, buf, CRC_BUF_SIZE)) > 0) {
    *crc = crc32(*crc, buf, count);
  }
  
  if (count < 0) {
    cal->err = CAL_ERR_IO;
    return 0;
  }

  return 1;
}
#undef CRC_BUF_SIZE

static int is_valid_crc(struct calfile *cal, uint32_t crc) {
  uint32_t l_crc;
  
  if (!calc_crc(cal, &l_crc))
    return 0;
  
  if (crc != l_crc) {
    cal->err = CAL_ERR_CHECKSUM;
    return 0;
  }

  return 1;
}

static int validate_and_version(struct calfile *cal) {
  uint32_t tmp;
  int ret = 1;
  
  /* read magic */
  ret = readu32(cal->fd, &tmp);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }
  if (tmp != CAL_MAGIC) {
    cal->err = CAL_ERR_MAGIC;
    return -1;
  }

  /* checksum */
  ret = readu32(cal->fd, &tmp);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }
  if (!is_valid_crc(cal, tmp))
    return -1;

  /* version */
  if (lseek(cal->fd, 8, SEEK_SET) != 8) {
    cal->err = CAL_ERR_IO;
    return -1;
  }
  ret = readu32(cal->fd, &tmp);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }

  if (tmp < 0 || tmp > CAL_VERSION_MAX) {
    cal->err = CAL_ERR_UNSUPPORTED_VERSION;
    return -1;
  }
  
  /* return the version */
  return tmp;
}

static int write_comm_header(struct calfile *cal) {
  int ret;

  /* write magic */
  ret = writeu32(cal->fd, CAL_MAGIC);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }

  /* write checksum place holder */
  ret = writeu32(cal->fd, 0);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }

  /* write version */
  cal->version = CAL_LIB_VERSION;
  ret = writeu32(cal->fd, cal->version);
  if (!ret) {
    cal->err = CAL_ERR_IO;
    return -1;
  }

  return 0;
}

/* static off_t filesize(struct calfile *cal) { */
/*   struct stat stat; */
/*   int ret; */

/*   ret = fstat(cal->fd, &stat); */
/*   if (ret == -1) { */
/*     cal->err = CAL_ERR_IO; */
/*     return -1; */
/*   } */

/*   /\* are we looking at a symlink by mistake? *\/ */
/*   if (S_ISLNK(stat.st_mode)) { */
/*     cal->err = CAL_ERR_SYMLINK; */
/*     return -1; */
/*   } */

/*   return stat.st_size; */
/* } */

void set_string(struct calfile *cal, const char *s) {
  strncpy(cal->string, s, CAL_MAX_STRING_SIZE);
  cal->string[CAL_MAX_STRING_SIZE-1] = '\0';
}

/*
 *   CAL_OPEN
 */

/* cal open for version 0 of the file */
static void cal_open_0(struct calfile *cal) {
  int ret;
  /*  off_t len; */

  switch (cal->mode) {
  case CAL_READ:
    /* read timestamp */
    if (lseek(cal->fd, 12, SEEK_SET) != 12) {
      cal->err = CAL_ERR_IO;
      return;
    }

    ret = readu32(cal->fd, &cal->timestamp);
    if (!ret) {
      cal->err = CAL_ERR_IO;
      return;
    }
    
    /* read extra data len */
    ret = readu32(cal->fd, &cal->extra_data_len);
    if (!ret) {
      cal->err = CAL_ERR_IO;
      return;
    }
    
    /* do we have extra data ? */
    cal->next_pair = cal->extra_data_len?CAL_HEADERSIZE_0:0;

    cal->next_data = cal->extra_data_len + CAL_HEADERSIZE_0;

    /* validate file length */
    /*
    len = filesize(cal);
    if (len != (CAL_DATASIZE_0 + cal->extra_data_len + CAL_HEADERSIZE_0)) {
      cal->err = CAL_ERR_FILESIZE;
      return;
    }
    */

    break;

  case CAL_CREATE:
    
    /* add the timestamp */
    if (lseek(cal->fd, 12, SEEK_SET) != 12) {
      cal->err = CAL_ERR_IO;
      return;
    }
    
    cal->timestamp = time(NULL);
    ret = writeu32(cal->fd, cal->timestamp);
    if (!ret) {
      cal->err = CAL_ERR_IO;
      return;
    }

    /* put a place holder for extra data len */
    ret = writeu32(cal->fd, 0);
    if (!ret) {
      cal->err = CAL_ERR_IO;
      return;
    }
    
    cal->extra_data_len = 0; /* no extra data, yet */

    break;

  default: /* should never happen */
    break;
  }
}

typedef void (*cal_open_fn)(struct calfile *cal);
static const cal_open_fn cal_open_array[] = {
  cal_open_0,
}; 

enum calerr cal_open(struct calfile *cal, const char *fname, enum calmode mode) {
  int option = 0;

  /* replace or create? */
  switch (mode) {
  case CAL_REPLACE:
    option = O_TRUNC;
    mode = CAL_CREATE;
    break;
  case CAL_CREATE:
    option = O_EXCL;
    break;
  case CAL_READ:
    break;
  default:
    cal->err = CAL_ERR_INVALID_ARG;
    return cal->err;
  }

  cal->err = CAL_ERR_NONE;
  cal->mode = mode;
  cal->filename = NULL;

  switch (mode) {
  case CAL_READ:
    cal->state = CAL_STATE_RDONLY;

    /* open the file */
    cal->fd = open(fname, O_RDONLY);
    if (cal->fd == -1) {
      cal->err = CAL_ERR_FILE_OPEN;
      break;
    }

    /* get the version and check crc */
    if ((cal->version = validate_and_version(cal)) == -1)
      break;
    
    break;

  case CAL_CREATE:
    cal->state = CAL_STATE_HEADER;

    /* open the file */
    cal->fd = open(fname, O_RDWR | O_CREAT | option, 0664);
    if (cal->fd == -1) {
      cal->err = CAL_ERR_FILE_OPEN;
      break;
    }

    /* save the filename */
    cal->filename = malloc(strlen(fname) + 1);
    strcpy(cal->filename, fname);

    /* put the beginning of the header on */
    if (write_comm_header(cal) == -1)
      break;

    break;

  default:
    cal->err = CAL_ERR_INVALID_ARG;
  }

  /* only the first 3 fields have been handled thus far */

  /* do version specific open */
  if (cal->err == CAL_ERR_NONE)
    cal_open_array[cal->version](cal);

  /* if something bad happened -- close the file and return the error */
  if (cal->err != CAL_ERR_NONE && cal->fd != -1)
    cal_close(cal);

  return cal->err;
}


/*
 *   CAL_CLOSE
 */

static void cal_close_0(struct calfile *cal) {
  /* off_t len; */
  uint32_t crc;

  switch(cal->mode) {
  case CAL_READ:
    /* only reading, just close it */
    break;

  case CAL_CREATE:
    /* has enough cal data been written ? */
    /*
    len = filesize(cal);
    if (len != (CAL_DATASIZE_0 + cal->extra_data_len + CAL_HEADERSIZE_0)) {
      cal->err = CAL_ERR_FILESIZE;
      return;
    }
    */

    /* write extra data size too */
    if (lseek(cal->fd, 16, SEEK_SET) != 16) {
      cal->err = CAL_ERR_IO;
      break;
    }

    if (!writeu32(cal->fd, cal->extra_data_len)) {
      cal->err = CAL_ERR_IO;
      break;
    }

    /* we need to update the crc */
    if (!calc_crc(cal, &crc))
      break;

    if (lseek(cal->fd, 4, SEEK_SET) != 4) {
      cal->err = CAL_ERR_IO;
      break;
    }
    
    if (!writeu32(cal->fd, crc)) {
      cal->err = CAL_ERR_IO;
      break;
    }
    
    /* actually close the file */
    break;
    
  default:
    cal->err = CAL_ERR_DAMAGED_STRUCT;
    return;
  }
  
  /* return to hard close */
}

typedef void (*cal_close_fn)(struct calfile *cal);
static const cal_close_fn cal_close_array[] = {
  cal_close_0,
};

enum calerr cal_close(struct calfile *cal) {
  if (cal->err == CAL_ERR_NONE) {

    if (cal->version < 0 || cal->version > CAL_VERSION_MAX) {
      cal->err = CAL_ERR_DAMAGED_STRUCT;
    }
    else {
      switch (cal->mode) {
      case CAL_READ:
      case CAL_CREATE:
	cal_close_array[cal->version](cal);
	break;
      default:
	cal->err = CAL_ERR_DAMAGED_STRUCT;
      }
    }
  }

  /* free filename and unlink file if something went wrong? */
  if (cal->filename) {
    if (cal->mode == CAL_CREATE && cal->err != CAL_ERR_NONE)
      unlink(cal->filename);
    free(cal->filename);
  }

  close(cal->fd);
  cal->fd = -1;

  return cal->err;
}

/*
 *   CAL_GETERRSTRING
 */

/* this shouldn't be version specific */
const char *cal_geterrstring(struct calfile *cal) {
  if (cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }

  /* give the error */
  set_string(cal, cal_err_strings[cal->err]);

  return cal->string;
}

/*
 *   CAL_RESETKEYENUM
 */

static void cal_resetkeyenum_0(struct calfile *cal) {
  if (cal->extra_data_len)
    cal->next_pair = CAL_HEADERSIZE_0;
  else
    cal->next_pair = 0; /* no extras */
}

typedef void (*cal_resetkeynum_fn)(struct calfile *cal);
static const cal_resetkeynum_fn cal_resetkeyenum_array[] = {
  cal_resetkeyenum_0,
};

enum calerr cal_resetkeyenum(struct calfile *cal) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal_resetkeyenum_array[cal->version](cal);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;
}

/*
 *   CAL_GETNEXTKEY
 */

static const char *cal_getnextkey_0(struct calfile *cal) {
  uint16_t keylen, valuelen;

  if (cal->next_pair) {
    /* move to next pair */
    if (lseek(cal->fd, cal->next_pair, SEEK_SET) != cal->next_pair) {
      cal->err = CAL_ERR_IO;
      return NULL;
    }

    /* read key */
    if (!readu16(cal->fd, &keylen)) {
      cal->err = CAL_ERR_IO;
      return NULL;
    }

    /* too long? */
    if (keylen > CAL_MAX_STRING_SIZE-1) {
      cal->err = CAL_ERR_STRING_TOO_LARGE;
      return NULL;
    }

    /* read the value len */
    if (!readu16(cal->fd, &valuelen)) {
      cal->err = CAL_ERR_IO;
      return NULL;
    }

    /* read the key */
    if (read(cal->fd, cal->string, keylen) != keylen) {
      cal->err = CAL_ERR_IO;
      return NULL;
    }
    /* null terminate */
    cal->string[keylen] = '\0';

    /* find the next pair */
    /* this the last one? */
    if ((cal->next_pair + keylen + valuelen + 4 - CAL_HEADERSIZE_0) >= (cal->extra_data_len)) {
      cal->next_pair = 0;
    }
    else {
      cal->next_pair += 4 + keylen + valuelen;
    }

    /* return the results */
    return cal->string;
  }
  else {
    /* no more left */
    return NULL;
  }
}

typedef const char *(*cal_getnextkey_fn)(struct calfile *cal);
static const cal_getnextkey_fn cal_getnextkey_array[] = {
  cal_getnextkey_0,
};

const char *cal_getnextkey(struct calfile *cal) {
  const char *result = NULL;
  
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    /* called with an error? */
    return NULL;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      result = cal_getnextkey_array[cal->version](cal);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }
  
  if (cal->err != CAL_ERR_NONE) {
    cal_close(cal);
    return NULL;
  }
  else
    return result;
}


/*
 *   CAL_GETVALUE
 */

static int getpair(struct calfile *cal, off_t pos, char *key, char *val, int max_len) {
  uint16_t keylen, vallen;

  if (lseek(cal->fd, pos, SEEK_SET) != pos) {
    cal->err = CAL_ERR_IO;
    return 0;
  }

  if (!readu16(cal->fd, &keylen)) {
    cal->err = CAL_ERR_IO;
    return 0;
  }

  if (!readu16(cal->fd, &vallen)) {
    cal->err = CAL_ERR_IO;
    return 0;
  }

  if (keylen > max_len-1 || vallen > max_len-1) {
    cal->err = CAL_ERR_STRING_TOO_LARGE;
    return 0;
  }

  if (read(cal->fd, key, keylen) != keylen) {
    cal->err = CAL_ERR_IO;
    return 0;
  }
  key[keylen] = '\0';

  if (read(cal->fd, val, vallen) != vallen) {
    cal->err = CAL_ERR_IO;
    return 0;
  }
  val[vallen] = '\0';

  return keylen + vallen + 4; /* how much data was read */
}

const char *cal_getvalue_0(struct calfile *cal, const char *t_key) {
  off_t pos;
  char key[CAL_MAX_STRING_SIZE], val[CAL_MAX_STRING_SIZE];
  int off;

  pos = CAL_HEADERSIZE_0;

  while (pos < CAL_HEADERSIZE_0 + cal->extra_data_len) {
    off = getpair(cal, pos, key, val, CAL_MAX_STRING_SIZE);
    if (off == 0)
      break;

    /* found it? */
    if (!strcmp(t_key, key)) {
      set_string(cal, val);
      return cal->string;
    }

    /* try again */
    pos += off;
  }

  return NULL;
}

typedef const char *(*cal_getvalue_fn)(struct calfile *cal, const char *key);
static const cal_getvalue_fn cal_getvalue_array[] = {
  cal_getvalue_0,
};

const char *cal_getvalue(struct calfile *cal, const char *key) {
  const char *result = NULL;
  
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return NULL;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      result = cal_getvalue_array[cal->version](cal, key);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }
  
  if (cal->err != CAL_ERR_NONE) {
    cal_close(cal);
    return NULL;
  }
  else
    return result;
}


/*
 *   CAL_SETVALUE
 */

static void cal_setvalue_0(struct calfile *cal, const char *key, const char *value) {
  uint16_t keylen, vallen;
  off_t pos;

  if (cal->state != CAL_STATE_HEADER) {
    cal->err = CAL_ERR_INVALID_CALL_ORDER;
    return;
  }
  
  keylen = strlen(key);
  vallen = strlen(value);
  pos = CAL_HEADERSIZE_0 + cal->extra_data_len;

  if (!writeu16(cal->fd, keylen)) {
    cal->err = CAL_ERR_IO;
    return;
  }

  if (!writeu16(cal->fd, vallen)) {
    cal->err = CAL_ERR_IO;
    return;
  }

  if (write(cal->fd, key, keylen) != keylen) {
    cal->err = CAL_ERR_IO;
    return;
  }
  
  if (write(cal->fd, value, vallen) != vallen) {
    cal->err = CAL_ERR_IO;
    return;
  }

  cal->extra_data_len += (keylen + vallen + 4);
}

typedef void (*cal_setvalue_fn)(struct calfile *cal, const char *key, const char *val);
static const cal_setvalue_fn cal_setvalue_array[] = {
  cal_setvalue_0,
};

enum calerr cal_setvalue(struct calfile *cal, const char *key, const char *value) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal->err = CAL_ERR_READONLY;
      break;
    case CAL_CREATE:
      cal_setvalue_array[cal->version](cal, key, value);
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;
}


/*
 *   CAL_RESETDATAENUM
 */

static void cal_resetdataenum_0(struct calfile *cal) {
  cal->next_data = CAL_HEADERSIZE_0 + cal->extra_data_len;
}

typedef void (*cal_resetdataenum_fn)(struct calfile *cal);
static const cal_resetdataenum_fn cal_resetdataenum_array[] = {
  cal_resetdataenum_0,
};

enum calerr cal_resetdataenum(struct calfile *cal) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal_resetdataenum_array[cal->version](cal);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;  
}

/*
 *   CAL_NEXTDATATYPE
 */

void cal_getnextdatatype_0(struct calfile *cal, uint32_t *type) {
  if (lseek(cal->fd, cal->next_data, SEEK_SET) != cal->next_data) {
    cal->err = CAL_ERR_IO;
    return;
  }

  if (!readu32(cal->fd, type)) {
    /* end of file */
    *type = 0;
    return;
  }
}

typedef void (*cal_getnextdatatype_fn)(struct calfile *cal, uint32_t *type);
static const cal_getnextdatatype_fn cal_getnextdatatype_array[] = {
  cal_getnextdatatype_0,
};

enum calerr cal_getnextdatatype(struct calfile *cal, uint32_t *type) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal_getnextdatatype_array[cal->version](cal, type);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;  
}


/*
 *   CAL_GETDATA
 */

static void cal_getdata_0(struct calfile *cal, uint32_t *data) {
  uint32_t type;
  uint32_t count;

  if (lseek(cal->fd, cal->next_data, SEEK_SET) != cal->next_data) {
    cal->err = CAL_ERR_IO;
    return;
  }

  if (!readu32(cal->fd, &type)) {
    cal->err = CAL_ERR_IO;
    return;
  }

  count = CALTYPE_COUNT(type);

  cal->next_data += 4 + count*4;

  if (!data)
    return;

  for (; count; count--, data++) {
    if (!readu32(cal->fd, data)) {
      cal->err = CAL_ERR_IO;
      return;
    }
  }
}

typedef void (*cal_getdata_fn)(struct calfile *cal, uint32_t *data);
static const cal_getdata_fn cal_getdata_array[] = {
  cal_getdata_0,
};

enum calerr cal_getdata(struct calfile *cal, uint32_t *data) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal_getdata_array[cal->version](cal, data);
      break;
    case CAL_CREATE:
      cal->err = CAL_ERR_WRITEONLY;
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;
}


/*
 *   CAL_ADDCAL
 */

void cal_adddata_0(struct calfile *cal, const uint32_t *data, uint32_t type) {
  int count;

  /* once we start adding cal the "extra data" (key/value pairs) cannot be added */
  switch (cal->state) {
  case CAL_STATE_RDONLY:
    cal->err = CAL_ERR_READONLY;
    return;
  case CAL_STATE_HEADER:
    cal->state = CAL_STATE_DATA;
  case CAL_STATE_DATA:
    break;
  default:
    cal->err = CAL_ERR_DAMAGED_STRUCT;
    return;
  }

  /* move to the end of the file */
  if (lseek(cal->fd, 0, SEEK_END) == -1) {
    cal->err = CAL_ERR_IO;
    return;
  }

  /* write type */
  if (!writeu32(cal->fd, type)) {
    cal->err = CAL_ERR_IO;
    return;
  }
  
  count = CALTYPE_COUNT(type);

  /* write the data */
  for (; count; count--) {
    if (!writeu32(cal->fd, *data++)) {
      cal->err = CAL_ERR_IO;
      return;
    }
  }
}

typedef void (*cal_adddata_fn)(struct calfile *cal, const uint32_t *data, uint32_t type);
static const cal_adddata_fn cal_adddata_array[] = {
  cal_adddata_0,
};

enum calerr cal_adddata(struct calfile *cal, const void *data, uint32_t type) {
  if (cal->version > CAL_VERSION_MAX ||
      cal->version < 0 ||
      cal->err > CAL_ERR_MAX ||
      cal->err < CAL_ERR_MIN) {
    cal->err = CAL_ERR_DAMAGED_STRUCT;
  }
  else if (cal->err != CAL_ERR_NONE) {
    return cal->err;
  }
  else {  
    switch (cal->mode) {
    case CAL_READ:
      cal->err = CAL_ERR_READONLY;
      break;
    case CAL_CREATE:
      cal_adddata_array[cal->version](cal, data, type);
      break;
    default:
      cal->err = CAL_ERR_DAMAGED_STRUCT;
      break;
    }
  }

  /* something bad happen? */
  if (cal->err != CAL_ERR_NONE)
    cal_close(cal);

  return cal->err;
}

