#ifndef _CALFILE_H
#define _CALFILE_H

#include <inttypes.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

/*
 * If a new version of the file format needs to be supported (checklist):
 *
 * 1. CAL_VERSION_MAX & CAL_LIB_VERSION should be incremented.
 * 2. CAL_HEADERSIZE_? and CAL_DATASIZE_? should be defined (where ? is the new version number).
 * 3. Add errors to enum calerr and cal_err_strings as appropriate.
 * 4. Add static functions ending in _? (open, close) to perform the new operations.
 * 5. Add these new functions to the const arrays index on version number.
 * 6. If new extern functions are needed, use the array system and create a 'not supported'
 *      error function for all earlier versions (or return a default value if that works).
 * 7. If new data is needed in calfile, put it in "private" pointer on open and destroy it
 *      on close.
 *
 */

#define CAL_MAX_STRING_SIZE 256
#define CAL_MAGIC 0xbabe4abe

/* these should always be the same... we should write the newest format */
#define CAL_VERSION_MAX 0
#define CAL_LIB_VERSION 0

/*
 * CALFILE VERSION 0 defines
 */
#define CAL_HEADERSIZE_0 20

#define CALTYPE(t,c) ((t) << 16 | (c))
#define CALTYPE_COUNT(t) ((t) & 0xFFFF)

#define CALBANK_SIZE 256
#define CALTYPE_CAL_65MV_DELTA CALTYPE(0x0001, 1)
#define CALTYPE_VREF           CALTYPE(0x0002, 1)
#define CALTYPE_FACT_GAIN_POS  CALTYPE(0x0003, CALBANK_SIZE)
#define CALTYPE_FACT_GAIN_NEG  CALTYPE(0x0004, CALBANK_SIZE)
#define CALTYPE_SELF_GAIN_POS  CALTYPE(0x0005, CALBANK_SIZE)
#define CALTYPE_SELF_GAIN_NEG  CALTYPE(0x0006, CALBANK_SIZE)
#define CALTYPE_FACT_OFFSET    CALTYPE(0x0007, CALBANK_SIZE)
#define CALTYPE_SELF_OFFSET    CALTYPE(0x0008, CALBANK_SIZE)
#define CALTYPE_USER_GAIN      CALTYPE(0x0009, CALBANK_SIZE)
#define CALTYPE_USER_OFFSET    CALTYPE(0x000A, CALBANK_SIZE)

/* library error codes */
enum calerr {
  CAL_ERR_MIN = 0,
  CAL_ERR_NONE = CAL_ERR_MIN,
  CAL_ERR_CHECKSUM,
  CAL_ERR_FILE_OPEN,
  CAL_ERR_INVALID_CALL_ORDER,
  CAL_ERR_UNSUPPORTED_CALL,
  CAL_ERR_READONLY,
  CAL_ERR_WRITEONLY,
  CAL_ERR_IO,
  CAL_ERR_INVALID_ARG,
  CAL_ERR_MAGIC,
  CAL_ERR_UNSUPPORTED_VERSION,
  CAL_ERR_DAMAGED_STRUCT,
  CAL_ERR_SYMLINK,
  CAL_ERR_FILESIZE,
  CAL_ERR_STRING_TOO_LARGE,
  CAL_ERR_MAX = CAL_ERR_STRING_TOO_LARGE,
};

/* this is used internally to keep track of where we are */
enum calstate {
  CAL_STATE_RDONLY,
  CAL_STATE_HEADER,
  CAL_STATE_DATA,
};

/* file open mode */
/* cal files are either READ_ONLY or WRITE_ONLY, never both */
enum calmode {
  CAL_READ,
  CAL_CREATE,
  CAL_REPLACE,
};

/* the main calfile struct */
/* THIS DATA SHOULD NOT BE ACCESSED DIRECTLY */
struct calfile {
  uint32_t version;
  uint32_t timestamp;
  uint32_t extra_data_len;  /* length of key/value block */
  enum calerr err;
  enum calstate state;
  enum calmode mode;
  off_t next_pair;  /* reading key/value pairs, 0 == END */
  off_t next_data;
  int fd;
  char string[CAL_MAX_STRING_SIZE];
  char *filename;
  void *ver_specific;  /* private -- not currently used */
};

/* **************************
 * main library interface
 ***************************/

/* open a file
 * 
 * side effects: checks checksum
 * mode: READ | CREATE
 * return: open error code
 */
extern enum calerr cal_open(struct calfile *cal, const char *fname, enum calmode mode);

/* close a file
 *
 * side effects: write header info (checksum, timestamp, etc)
 * mode: READ | CREATE
 * return: close errors
 */
extern enum calerr cal_close(struct calfile *cal);

/* get error code
 *
 * mode: READ | CREATE
 * return: last error
 */
#define cal_geterr(c) ((const enum calerr)(c)->err)

/* get error string
 *
 * mode: READ | CREATE
 * return: const string of error message
 */
extern const char *cal_geterrstring(struct calfile *cal);

/* get this file's version
 *
 * mode: READ | CREATE
 * return: file version
 */
#define cal_getversion(c) ((const uint32_t)(c)->version)

/* get this file's timestamp
 *
 * mode: READ | CREATE
 * return: timestamp (epoch)
 */
#define cal_gettimestamp(c) ((const uint32_t)(c)->timestamp)

/* reset key enum
 *
 * NOTE: use this to reset before looping on cal_getnextkey to list all the keys in the file
 * mode: READ
 * return: error if one
 */
extern enum calerr cal_resetkeyenum(struct calfile *cal);

/* gets the next key
 *
 * mode: READ
 * return: pointer to a shared (gets clobbered on the next call) buffer
 *    or NULL if no more are left
 */
extern const char *cal_getnextkey(struct calfile *cal);

/* gets a value for a key/value pair
 *
 * mode: READ
 * return: pointer to a shared (gets clobbered on the next call) buffer
 *    or NULL if this key does not exist
 */
extern const char *cal_getvalue(struct calfile *cal, const char *key);

/* add a key/value pair to a CREATE file
 *
 * mode: CREATE
 * NOTE: this can not be called after any cal_addcdata calls.
 * return: error if one
 */
extern enum calerr cal_setvalue(struct calfile *cal, const char *key, const char *value);

/* reset the data enum
 *
 * return: error if one
 */
extern enum calerr cal_resetdataenum(struct calfile *cal);

/* type of the next data block
 *
 * return (in type)
 *          0  -- EOF
 *         !0 -- type
 */
extern enum calerr cal_getnextdatatype(struct calfile *cal, uint32_t *type);

/* retrieve (next) cal data
 *
 * mode: READ
 * NOTE: if data=NULL, it just skips this block
 */
extern enum calerr cal_getdata(struct calfile *cal, uint32_t *data);

/* add cal data to a added to the end of cal file
 *
 * NOTE: calling this ends the ability to add key/values pairs
 * mode: CREATE
 * return: error if one
 */
extern enum calerr cal_adddata(struct calfile *cal, const void *data, uint32_t type);

#endif
