#include "calfile.h"
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

float cal_65mv_delta;
float vref;
float fact_gain_pos[CALBANK_SIZE];
float fact_gain_neg[CALBANK_SIZE];
float fact_offset[CALBANK_SIZE];
float self_gain_pos[CALBANK_SIZE];
float self_gain_neg[CALBANK_SIZE];
float self_offset[CALBANK_SIZE];
float user_gain[CALBANK_SIZE];
float user_offset[CALBANK_SIZE];
int cal_65mv_delta_en = 0;
int vref_en = 0;
int fact_gain_pos_en = 0;
int fact_gain_neg_en = 0;
int fact_offset_en = 0;
int self_gain_pos_en = 0;
int self_gain_neg_en = 0;
int self_offset_en = 0;
int user_gain_en = 0;
int user_offset_en = 0;

/* flags */
int fact_def = 0;
int self_def = 0;
int user_def = 0;
int verbose = 0;
int header = 0;
const char *binout = NULL;
const char *txtout = NULL;
const char *infile = NULL;

void die(const char *msg) {
  fprintf(stderr, "die: %s\n", msg);
  exit(1);
}

void usage(const char *progname) {
  fprintf(stderr, "Usage: %s [OPTIONS] [INPUT FILE]\n", progname);
  fprintf(stderr, "Manipulates cal files\n");
  fprintf(stderr, "\t-d         use factory defaults rather than an input file\n");
  fprintf(stderr, "\t-s         use self defaults rather than an input file\n");
  fprintf(stderr, "\t-u         use user defaults rather than an input file\n");
  fprintf(stderr, "\t-b <file>  binary output format to file\n");
  fprintf(stderr, "\t-f <file>  text output to specified file\n");
  fprintf(stderr, "\t-h         output file header info\n");
  exit(0);
}

void setfactdefault(void);
void setselfdefault(void);
void setuserdefault(void);
int readfile(const char *fname);
int readtxt(const char *fname);
int readbin(const char *fname);
int dumptxt(void);
int dumpbin(const char *fname);

int main(int argc, char *argv[]) {
  int c;
  char optstr[] = "dhb:f:sum:";

  while ((c = getopt(argc, argv, optstr)) != -1) {
    switch(c) {
    case 'd': fact_def = 1; break;
    case 's': self_def = 1; break;
    case 'u': user_def = 1; break;
    case 'h': header = 1; break;
    case 'b': binout = optarg; break;
    case 'f': txtout = optarg; break;
    default:
      usage(argv[0]);
    }
  }

  /* we need at least 1 source of input! (file/defaults) */
  if (((fact_def || self_def || user_def) && optind != argc) ||
      (!(fact_def || self_def || user_def) && optind+1 != argc))
    usage(argv[0]);

  /* handle inputs */
  if (fact_def || self_def || user_def) {
    if (fact_def)
      setfactdefault();
    if (self_def)
      setselfdefault();
    if (user_def)
      setuserdefault();
  }
  else {
    infile = argv[optind];
    if (!readfile(infile))
      die("could not identify input file as a valid binary or text file");
  }

  if (!header) {
    /* change output if necessary */
    if (txtout) {
      fclose(stdout);
      stdout = fopen(txtout, "w");
      if (!stdout)
	die("failed to open output text file");
    }
    
    if (binout) {
      if (!dumpbin(binout))
	die("binary output failed");
    }
    
    if (!binout || txtout) {
      if (!dumptxt())
	die("text output failed");
      fclose(stdout);
    }
  }
  else {
    printf("FIELDS:\n");
#define FIELD(a) if (a##_en) printf("> " #a "\n")
    FIELD(cal_65mv_delta);
    FIELD(vref);
    FIELD(fact_gain_pos);
    FIELD(fact_gain_neg);
    FIELD(fact_offset);
    FIELD(self_gain_pos);
    FIELD(self_gain_neg);
    FIELD(self_offset);
    FIELD(user_gain);
    FIELD(user_offset);
#undef FIELD
  }

  return 0;
}

void setfactdefault(void) {
  int i;

  cal_65mv_delta = 62.16f;
  cal_65mv_delta_en = 1;
  vref = 4.086;
  vref_en = 1;

  for (i = 0; i < CALBANK_SIZE; i++) {
    fact_gain_pos[i] = 1.0f;
    fact_gain_neg[i] = 1.0f;
  }
  fact_gain_pos_en = 1;
  fact_gain_neg_en = 1;
 
  for (i = 0; i < CALBANK_SIZE; i++)
    fact_offset[i] = 0.0f;
  fact_offset_en = 1;
}

void setselfdefault(void) {
  int i;

  for (i = 0; i < CALBANK_SIZE; i++) {
    self_gain_pos[i] = 1.0f;
    self_gain_neg[i] = 1.0f;
  }
  self_gain_pos_en = 1;
  self_gain_neg_en = 1;
 
  for (i = 0; i < CALBANK_SIZE; i++)
    self_offset[i] = 0.0f;
  self_offset_en = 1;
}

void setuserdefault(void) {
  int i;

  for (i = 0; i < CALBANK_SIZE; i++)
    user_gain[i] = 1.0f;
  user_gain_en = 1;
 
  for (i = 0; i < CALBANK_SIZE; i++)
    user_offset[i] = 0.0f;
  user_offset_en = 1;
}

int readfile(const char *fname) {
  switch (readbin(fname)) {
  case 1:
    return 1;
  case 0:
    return readtxt(fname);
  default:
    return 0;
  }
}

int readtxt(const char *fname) {
  FILE *fp;
  char key[128];
  char buf[256];
  int ind;
  float val;
  int ret;

  fp = fopen(fname, "r");
  if (!fp) {
    perror("error opening input text file");
    return 0;
  }

  while (fgets(buf, 256, fp)) {

    if (strchr(buf, '[')) {
      ret = sscanf(buf, "%[^[][%i]=%f\n", key, &ind, &val);
      if (ret != 3) {
	fprintf(stderr, "unabled to parse: %s", buf);
	fclose(fp);
	return 0;
      }
      if (ind < 0 || ind > 255) {
	fprintf(stderr, "index too large in: %s", buf);
	fclose(fp);
	return 0;
      }
      
      if      (!strcmp(key, "FACT_GAIN_POS")) {
	fact_gain_pos[ind] = val;
	fact_gain_pos_en = 1;
      }
      else if (!strcmp(key, "FACT_GAIN_NEG")) {
	fact_gain_neg[ind] = val;
	fact_gain_neg_en = 1;
      }
      else if (!strcmp(key, "FACT_OFFSET")) {
	fact_offset[ind] = val;
	fact_offset_en = 1;
      }
      else if (!strcmp(key, "SELF_GAIN_POS")) {
	self_gain_pos[ind] = val;
	self_gain_pos_en = 1;
      }
      else if (!strcmp(key, "SELF_GAIN_NEG")) {
	self_gain_neg[ind] = val;
	self_gain_neg_en = 1;
      }
      else if (!strcmp(key, "SELF_OFFSET")) {
	self_offset[ind] = val;
	self_offset_en = 1;
      }
      else if (!strcmp(key, "USER_GAIN")) {
	user_gain[ind] = val;
	user_gain_en = 1;
      }
      else if (!strcmp(key, "USER_OFFSET")) {
	user_offset[ind] = val;
	user_offset_en = 1;
      }
      else {
	fprintf(stderr, "unknown key: %s", key);
	fclose(fp);
	return 0;
      }
    }
    else {
      ret = sscanf(buf, "%[^=]=%f\n", key, &val);
      if (ret != 2) {
	fprintf(stderr, "unabled to parse: %s", buf);
	fclose(fp);
	return 0;
      }
      
      if      (!strcmp(key, "CAL_65MV_DELTA")) {
	cal_65mv_delta = val;
	cal_65mv_delta_en = 1;
      }
      else if (!strcmp(key, "VREF")) {
	vref = val;
	vref_en = 1;
      }
      else {
	fprintf(stderr, "unknown key: %s", key);
	fclose(fp);
	return 0;
      }
    }

  }

  fclose(fp);
  return 1;
}

int readbin(const char *fname) {
  struct calfile cal;
  time_t t;
  const char *key;
  uint32_t type;

  /* print binary file header to stderr */
  if (cal_open(&cal, fname, CAL_READ) != CAL_ERR_NONE)
    return 0;

  /* dump header output */
  if (header) {
    printf("HEADER:\n");
    printf("> version: %i\n", cal_getversion(&cal));
    t = cal_gettimestamp(&cal);
    printf("> timestamp: %s", ctime(&t));
    cal_resetkeyenum(&cal);

    while ((key = cal_getnextkey(&cal))) {
      printf("> %s: ", key);
      printf("%s\n", cal_getvalue(&cal, key));
    }
    if (cal_geterr(&cal) != CAL_ERR_NONE) {
      fprintf(stderr, "error in binary cal file: %s\n", cal_geterrstring(&cal));
      return -1;
    }
  }

  /* read the data blocks */
  cal_resetdataenum(&cal);
  while (1) {
    cal_getnextdatatype(&cal, &type);
    if (!type)
      break;

    switch (type) {
    case CALTYPE_CAL_65MV_DELTA:
      cal_getdata(&cal, (uint32_t*)&cal_65mv_delta);
      cal_65mv_delta_en = 1;
      break;
    case CALTYPE_VREF:
      cal_getdata(&cal, (uint32_t*)&vref);
      vref_en = 1;
      break;
    case CALTYPE_FACT_GAIN_POS:
      cal_getdata(&cal, (uint32_t*)fact_gain_pos);
      fact_gain_pos_en = 1;
      break;
    case CALTYPE_FACT_GAIN_NEG:
      cal_getdata(&cal, (uint32_t*)fact_gain_neg);
      fact_gain_neg_en = 1;
      break;
    case CALTYPE_FACT_OFFSET:
      cal_getdata(&cal, (uint32_t*)&fact_offset);
      fact_offset_en = 1;
      break;
    case CALTYPE_SELF_GAIN_POS:
      cal_getdata(&cal, (uint32_t*)self_gain_pos);
      self_gain_pos_en = 1;
      break;
    case CALTYPE_SELF_GAIN_NEG:
      cal_getdata(&cal, (uint32_t*)self_gain_neg);
      self_gain_neg_en = 1;
      break;
    case CALTYPE_SELF_OFFSET:
      cal_getdata(&cal, (uint32_t*)self_offset);
      self_offset_en = 1;
      break;
    case CALTYPE_USER_GAIN:
      cal_getdata(&cal, (uint32_t*)user_gain);
      user_gain_en = 1;
      break;
    case CALTYPE_USER_OFFSET:
      cal_getdata(&cal, (uint32_t*)user_offset);
      user_offset_en = 1;
      break;
    default:
      fprintf(stderr, "unexpected data block type in cal file\n");
      cal_close(&cal);
      return -1;
    }
  }

  if (cal_geterr(&cal) != CAL_ERR_NONE) {
    fprintf(stderr, "error in binary cal file: %s\n", cal_geterrstring(&cal));
    return -1;
  }

  cal_close(&cal);
  return 1;
}

int dumptxt(void) {
  int i;

  if (cal_65mv_delta_en)
    printf("CAL_65MV_DELTA=%+15.15e\n", cal_65mv_delta);

  if (vref_en)
    printf("VREF=%+15.15e\n", vref);

  if (fact_gain_pos_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("FACT_GAIN_POS[%3i]=%+15.15e\n", i, fact_gain_pos[i]);

  if (fact_gain_neg_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("FACT_GAIN_NEG[%3i]=%+15.15e\n", i, fact_gain_neg[i]);

  if (fact_offset_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("FACT_OFFSET[%3i]=%+15.15e\n", i, fact_offset[i]);

  if (self_gain_pos_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("SELF_GAIN_POS[%3i]=%+15.15e\n", i, self_gain_pos[i]);

  if (self_gain_neg_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("SELF_GAIN_NEG[%3i]=%+15.15e\n", i, self_gain_neg[i]);

  if (self_offset_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("SELF_OFFSET[%3i]=%+15.15e\n", i, self_offset[i]);

  if (user_gain_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("USER_GAIN[%3i]=%+15.15e\n", i, user_gain[i]);

  if (user_offset_en)
    for (i = 0; i < CALBANK_SIZE; i++)
      printf("USER_OFFSET[%3i]=%+15.15e\n", i, user_offset[i]);

  return 1;
}

int dumpbin(const char *fname) {
  struct calfile cal;

  if (cal_open(&cal, fname, CAL_CREATE) != CAL_ERR_NONE) {
    fprintf(stderr, "failed to open binary output: %s\n", cal_geterrstring(&cal));
    return 0;
  }

  cal_setvalue(&cal, "vender", "VXI Technology, Inc.");
  cal_setvalue(&cal, "device", "EX1048");
  cal_setvalue(&cal, "rev", "1");

  if (cal_65mv_delta_en)
    cal_adddata(&cal, &cal_65mv_delta, CALTYPE_CAL_65MV_DELTA);

  if (vref_en)
    cal_adddata(&cal, &vref, CALTYPE_VREF);
  
  if (fact_gain_pos_en)
    cal_adddata(&cal, fact_gain_pos, CALTYPE_FACT_GAIN_POS);

  if (fact_gain_neg_en)
    cal_adddata(&cal, fact_gain_neg, CALTYPE_FACT_GAIN_NEG);

  if (fact_offset_en)
    cal_adddata(&cal, fact_offset, CALTYPE_FACT_OFFSET);

  if (self_gain_pos_en)
    cal_adddata(&cal, self_gain_pos, CALTYPE_SELF_GAIN_POS);

  if (self_gain_neg_en)
    cal_adddata(&cal, self_gain_neg, CALTYPE_SELF_GAIN_NEG);

  if (self_offset_en)
    cal_adddata(&cal, self_offset, CALTYPE_SELF_OFFSET);

  if (user_gain_en)
    cal_adddata(&cal, user_gain, CALTYPE_USER_GAIN);

  if (user_offset_en)
    cal_adddata(&cal, user_offset, CALTYPE_USER_OFFSET);

  if (cal_close(&cal) != CAL_ERR_NONE) {
    fprintf(stderr, "binary cal output error: %s\n", cal_geterrstring(&cal));
    return 0;
  }

  return 1;
}
