#include "calfile.h"
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <acqmain.h>

static int verbose;
static int fd = -1;

#define NOMINAL_ZERO 0.0f
#define NOMINAL_15X_GAIN 0.006984f
#define NOMINAL_1X_GAIN 0.010476f
#define NOMINAL_2_DIV_3 0.666667f
#define NOMINAL_ONE 1.0f

#define DEFAULT_65MV_DELTA 0.06216f
#define DEFAULT_VREF 4.086f

#define U32TOFLOAT(a) (*(float*)&(a))
#define FLOATTOU32(a) (*(__u32*)&(a))

void usage(const char *pname) {
  fprintf(stderr, "Usage: %s [options] [cal file]\n", pname);
  fprintf(stderr, "\t when [cal file] is present, the fpga is loaded with it's data\n");
  fprintf(stderr, "\t when [cal file] is not present, the fpga cal state is printed/filled with defaults\n");
  fprintf(stderr, "\t-v  verbose\n");
  fprintf(stderr, "\t-d  load all fields with default values\n");

  fprintf(stderr, "\t-n  load all fields with nominal values\n");
  if (fd != -1)
    close(fd);
  exit(1);
}

int get_cal(void) {
  struct cal_bank bank;
  float val;
  int ret = 0, i;

#define RET(a) if (ret) { \
                 fprintf(stderr, a); \
                 return 1; \
               }
 
  ret = ioctl(fd, CAL_GET_CAL_65MV_DELTA, &val);
  RET("ioctl: CAL_GET_CAL_65MV_DELTA failed\n");
  printf("CAL_65MV_DELTA: %f\n", val);

  ret = ioctl(fd, CAL_GET_VREF, &val);
  RET("ioctl: CAL_GET_VREF failed\n");
  printf("VREF: %f\n", val);

  bank.field = CALBANK_FACT_GAIN_POS;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (FACT_GAIN_POS) failed\n");
  printf("FACT_GAIN_POS: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_FACT_GAIN_NEG;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (FACT_GAIN_NEG) failed\n");
  printf("FACT_GAIN_NEG: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_SELF_GAIN_POS;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (SELF_GAIN_POS) failed\n");
  printf("SELF_GAIN_POS: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_SELF_GAIN_NEG;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (SELF_GAIN_NEG) failed\n");
  printf("SELF_GAIN_NEG: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_FACT_OFFSET;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (FACT_OFFSET) failed\n");
  printf("FACT_OFFSET: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_SELF_OFFSET;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (SELF_OFFSET) failed\n");
  printf("SELF_OFFSET: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_USER_GAIN;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (USER_GAIN) failed\n");
  printf("USER_GAIN: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");

  bank.field = CALBANK_USER_OFFSET;
  ret = ioctl(fd, CAL_GET_BANK, &bank);
  RET("ioctl: CAL_GET_BANK (USER_OFFSET) failed\n");
  printf("USER_OFFSET: ");
  for (i = 0; i < CALBANK_SIZE; i++)
    printf("%1.4f ", U32TOFLOAT(bank.data[i]));
  printf("\n");


#undef RET
  return ret;
}

#define RET(a) if (ret) { \
                 fprintf(stderr, a); \
                 return 1; \
               }
 
int set_def(void) {
  struct cal_bank bank;
  float val;
  int ret = 0, i;

  val = DEFAULT_65MV_DELTA; /* default value */
  ret = ioctl(fd, CAL_SET_CAL_65MV_DELTA, &val);
  RET("ioctl: CAL_SET_CAL_65MV_DELTA failed\n");

  val = DEFAULT_VREF;
  ret = ioctl(fd, CAL_SET_VREF, &val);
  RET("ioctl: CAL_SET_VREF failed\n");

  /* do all Zeros (offsets) */
  for (i = 0; i < CALBANK_SIZE; i++)
    U32TOFLOAT(bank.data[i]) = 0.0f;

  bank.field = CALBANK_FACT_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_OFFSET) failed\n");

  bank.field = CALBANK_SELF_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_OFFSET) failed\n");

  bank.field = CALBANK_USER_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (USER_OFFSET) failed\n");

  /* do all Ones (gains) */
  for (i = 0; i < CALBANK_SIZE; i++)
    U32TOFLOAT(bank.data[i]) = 1.0f;

  bank.field = CALBANK_FACT_GAIN_POS;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_GAIN_POS) failed\n");

  bank.field = CALBANK_FACT_GAIN_NEG;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_GAIN_NEG) failed\n");

  bank.field = CALBANK_SELF_GAIN_POS;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_GAIN_POS) failed\n");

  bank.field = CALBANK_SELF_GAIN_NEG;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_GAIN_NEG) failed\n");

  bank.field = CALBANK_USER_GAIN;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (USER_GAIN) failed\n");

  if (verbose)
    printf("successfully loaded defaults.\n");

  return ret;
}

int set_nom(void) {
  struct cal_bank bank;
  float val;
  int ret = 0, i;

  val = DEFAULT_65MV_DELTA; /* default value */
  ret = ioctl(fd, CAL_SET_CAL_65MV_DELTA, &val);
  RET("ioctl: CAL_SET_CAL_65MV_DELTA failed\n");

  val = DEFAULT_VREF;
  ret = ioctl(fd, CAL_SET_VREF, &val);
  RET("ioctl: CAL_SET_VREF failed\n");

  /* do all Zeros (offsets) */
  for (i = 0; i < CALBANK_SIZE; i++)
    U32TOFLOAT(bank.data[i]) = NOMINAL_ZERO;

  bank.field = CALBANK_FACT_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_OFFSET) failed\n");

  bank.field = CALBANK_SELF_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_OFFSET) failed\n");

  bank.field = CALBANK_USER_OFFSET;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (USER_OFFSET) failed\n");

  for (i = 0; i < 48; i++) {
    U32TOFLOAT(bank.data[i      ]) = NOMINAL_15X_GAIN;
    U32TOFLOAT(bank.data[i + 64 ]) = NOMINAL_15X_GAIN;
    U32TOFLOAT(bank.data[i + 128]) = NOMINAL_1X_GAIN;
    U32TOFLOAT(bank.data[i + 192]) = NOMINAL_1X_GAIN;
  }
  for (i = 48; i < 64; i++) {
    U32TOFLOAT(bank.data[i]) = NOMINAL_2_DIV_3;
  }

  bank.field = CALBANK_FACT_GAIN_POS;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_GAIN_POS) failed\n");

  bank.field = CALBANK_FACT_GAIN_NEG;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (FACT_GAIN_NEG) failed\n");

  for (i = 0; i < CALBANK_SIZE; i++)
    U32TOFLOAT(bank.data[i]) = NOMINAL_ONE;

  bank.field = CALBANK_SELF_GAIN_POS;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_GAIN_POS) failed\n");

  bank.field = CALBANK_SELF_GAIN_NEG;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (SELF_GAIN_NEG) failed\n");

  bank.field = CALBANK_USER_GAIN;
  ret = ioctl(fd, CAL_SET_BANK, &bank);
  RET("ioctl: CAL_SET_BANK (USER_GAIN) failed\n");

  if (verbose)
    printf("successfully loaded nominals.\n");

  return ret;
}

#undef RET

int set_cal(const char *fname) {
  struct calfile cal;
  struct cal_bank bank;
  float val;
  uint32_t type;
  int ret = 0;

  /* read input */
  if (cal_open(&cal, fname, CAL_READ) != CAL_ERR_NONE) {
    fprintf(stderr, "attempt to open %s failed with error: %s\n", fname, cal_geterrstring(&cal));
    return 1;
  }

  /* print header if verbose */
  if (verbose) {
    uint32_t timestamp;
    const char *key;

    printf("version: %i\n", cal_getversion(&cal));
    timestamp = cal_gettimestamp(&cal);
    printf("timestamp: %s", ctime((time_t*)&timestamp));
    cal_resetkeyenum(&cal);
    
    while ((key = cal_getnextkey(&cal))) {
      printf("%s: ", key);
      printf("%s\n", cal_getvalue(&cal, key));
    }

    if (cal_geterr(&cal) != CAL_ERR_NONE) {
      fprintf(stderr, "error reading cal header: %s\n", cal_geterrstring(&cal));
      return 1;
    }
  }

#define GETDATA(a) if (cal_getdata(&cal, (uint32_t*)(a)) != CAL_ERR_NONE) { \
                     fprintf(stderr, "cal data error: %s\n", cal_geterrstring(&cal)); \
                     return 1; \
                   }
#define MSG(a) ((a)?"FAILED":"OK")

  /* start reading data */
  cal_resetdataenum(&cal);
  do {
    /* error? */
    if (cal_getnextdatatype(&cal, &type) != CAL_ERR_NONE)
      break;
    if (!type)
      break;
    
    switch (type) {
    case CALTYPE_CAL_65MV_DELTA:
      GETDATA(&val);
      ret = ioctl(fd, CAL_SET_CAL_65MV_DELTA, &val);
      if (verbose)
	printf("CAL_65MV_DELTA: (%f) %s\n", val, MSG(ret));
      break;
    case CALTYPE_VREF:      
      GETDATA(&val);
      ret = ioctl(fd, CAL_SET_VREF, &val);
      if (verbose)
	printf("VREF: (%f) %s\n", val, MSG(ret));
      break;
    case CALTYPE_FACT_GAIN_POS:
      GETDATA(&bank.data);
      bank.field = CALBANK_FACT_GAIN_POS;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("FACT_GAIN_POS: %s\n", MSG(ret));
      break;
    case CALTYPE_FACT_GAIN_NEG:
      GETDATA(&bank.data);
      bank.field = CALBANK_FACT_GAIN_NEG;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("FACT_GAIN_NEG: %s\n", MSG(ret));
      break;
    case CALTYPE_SELF_GAIN_POS:
      GETDATA(&bank.data);
      bank.field = CALBANK_SELF_GAIN_POS;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("SELF_GAIN_POS: %s\n", MSG(ret));
      break;
    case CALTYPE_SELF_GAIN_NEG:
      GETDATA(&bank.data);
      bank.field = CALBANK_SELF_GAIN_NEG;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("SELF_GAIN_NEG: %s\n", MSG(ret));
      break;
    case CALTYPE_FACT_OFFSET:
      GETDATA(&bank.data);
      bank.field = CALBANK_FACT_OFFSET;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("FACT_OFFSET: %s\n", MSG(ret));
      break;
    case CALTYPE_SELF_OFFSET:
      GETDATA(&bank.data);
      bank.field = CALBANK_SELF_OFFSET;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("SELF_OFFSET: %s\n", MSG(ret));
      break;
    case CALTYPE_USER_GAIN:
      GETDATA(&bank.data);
      bank.field = CALBANK_USER_GAIN;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("USER_GAIN: %s\n", MSG(ret));
      break;
    case CALTYPE_USER_OFFSET:
      GETDATA(&bank.data);
      bank.field = CALBANK_USER_OFFSET;
      ret = ioctl(fd, CAL_SET_BANK, &bank);
      if (verbose)
	printf("USER_OFFSET: %s\n", MSG(ret));
      break;
    }
  } while (!ret);
#undef GETDATA
#undef MSG

  cal_close(&cal);

  if (cal_geterr(&cal) != CAL_ERR_NONE) {
    fprintf(stderr, "cal file error: %s\n", cal_geterrstring(&cal));
    return 1;
  }

  return ret;
}

int main(int argc, char *argv[]) {
  char optstr[] = "vdn";
  int c;
  int ret = 0;
  int defaults = 0;
  int nominals = 0;

  /* parse opts */
  while ((c = getopt(argc, argv, optstr)) != -1) {
    switch (c) {
    case 'v': verbose = 1; break;
    case 'd': defaults = 1; break;
    case 'n': nominals = 1; break;
    default:
      usage(argv[0]);
    }
  }

  fd = open("/dev/acqcal0", O_RDWR);
  if (fd == -1) {
    fprintf(stderr, "failed to open device: /dev/acqcal0\n");
    return 1;
  }

  /* read/write/err */
  if (optind+1 == argc)
    ret = set_cal(argv[optind]);
  else if (optind == argc) {
    if (defaults)
      ret = set_def();
    else if (nominals)
      ret = set_nom();
    else
      ret = get_cal();
  }
  else
    usage(argv[0]);

  if (close(fd))
    perror("failed to close device: /dev/acqcal0");

  return ret;
}
