/*
  dmmlib.c
  
  DMM interface library
*/

#include <stdio.h>
#include <unistd.h>

#include "dmmlib.h"
#include "dmmdevices.h"


/*
  dmm_get_device_count
*/
int dmm_get_device_count (void) {
  return (DMM_DEVICE_MAX - DMM_DEVICE_MIN + 1);
}

/*
  dmm_get_device_name
*/
const char * dmm_get_device_name (int index) {
  /* convert the index to our internal index
     reall, DMM_DEVICE_MIN is always going to be 0; this is just here for completeness */
  index += DMM_DEVICE_MIN;
  
  if (index < DMM_DEVICE_MIN || index > DMM_DEVICE_MAX) {
    return NULL;
  }
  
  return g_devices[index].name;
}


/*
  dmm_init_start
*/

int dmm_init_start(const struct dmm_params **params, struct dmm_handle **dmm, struct device **device)
{
   /* check the dmm_info arguments */
  if ((*params)->device < DMM_DEVICE_MIN || 
      (*params)->device > DMM_DEVICE_MAX || 
      NULL == (*params)->host ||
      NULL == (*params)->gpib_addr) {
    return 1;
  }
  
  //fprintf( stderr, "passed argument test\n");
  
  /* allocate the dmm */
  *dmm = (struct dmm_handle *)malloc(sizeof(struct dmm_handle));
  if (NULL == *dmm) {
    return 1;
  }
  
  //fprintf(stderr, "DMM allocation OK\n");
  
  /* allocate the scpilib handle */
  (*dmm)->scpi = scpi_create();
  if (NULL == (*dmm)->scpi) {
    free(*dmm);
    return 1;
  }
  
  //fprintf(stderr,"SCPIlib OK\n");
  
  /* set misc. fields on the dmm handle */
 (*dmm)->device = (*params)->device;
  
  /* connect to the device */
  if (SCPI_ERROR_SUCCESS != scpi_connect((*dmm)->scpi, (*params)->host, (*params)->gpib_addr)) {
    scpi_destroy((*dmm)->scpi);
    free(*dmm);
    //fprintf(stderr,"failed to connect\n");
    return 1;
  }

  //fprintf(stderr, "Connect to Device OK\n");

  *device = &g_devices[(*dmm)->device];
  //fprintf(stderr, "DMM ID: %i\n",(*dmm)->device);
  if(*device==NULL)
  {
    //fprintf(stderr, "Failed to assign device\n");
    return 1;
  }
  return 0;
}


/*
  dmm_init
*/
struct dmm_handle * dmm_volt_init (const struct dmm_params *params) {
  struct dmm_handle *dmm;
  int i;
  struct device *device;
  int err;
  dmm = NULL;
  device = NULL;
  
  err = dmm_init_start(&params,&dmm,&device);
  if(err) {
  //fprintf(stderr,"DMM init start=%i\n",err);
  return NULL;
  }
   
  /* write out the list of commands needed to initialize the device */
  for (i=0; NULL != device->cmds->volt_init[i]; i++) {
    if (SCPI_ERROR_SUCCESS != (err = scpi_write(dmm->scpi, device->cmds->volt_init[i]))) {
      //printf("failed to send an init command: %s error=%i\n", device->cmds->init[i], err);
      return NULL;
    }
    /*sleep(2);*/
  }

  return dmm;
}

struct dmm_handle * dmm_res_init (const struct dmm_params *params) {
  struct dmm_handle *dmm;
  int i;
  struct device *device;
  int err;
  dmm = NULL;
  device = NULL;
  
  err = dmm_init_start(&params,&dmm,&device);
  if(err) {
  //fprintf(stderr,"DMM init start=%i\n",err);
  return NULL;
  }
   
  /* write out the list of commands needed to initialize the device */
  for (i=0; NULL != device->cmds->res_init[i]; i++) {
    if (SCPI_ERROR_SUCCESS != (err = scpi_write(dmm->scpi, device->cmds->res_init[i]))) {
      //printf("failed to send an init command: %s error=%i\n", device->cmds->init[i], err);
      return NULL;
    }
    /*sleep(2);*/
  }

  return dmm;
}


/*
  dmm_set_range_init
*/

int dmm_set_range_init(struct dmm_handle *handle,struct device **device) {
  *device = &g_devices[handle->device];
  if(*device==NULL) return 1;
  else return 0;
}


/*
  dmm_set_range_finish
*/
int dmm_set_range_finish(struct dmm_handle *handle,char** buf)
{
  int error= scpi_write(handle->scpi, *buf);
  if (error!=SCPI_ERROR_SUCCESS) {
    //fprintf(stderr,"error = %i\n",error);
    free(*buf);
    return 1;
  }
  else return 0;
}

/*
  dmm_set_range
*/

int dmm_set_volt_range (struct dmm_handle *handle, double value) {
  char *buf; /* buffer for the SCPI command */
  int buf_len;
  struct device *device;
  
  device=NULL;
 
  if(dmm_set_range_init( handle , &device ) != 0) {
     // fprintf(stderr,"Range init failed\n");
      return -1;
  }
  
  /* the buffer needs to contain:
     "<device's RANGE command> <range>\n\0<16 bytes of fudge>"
     where <range> is a float of the form %8.8f, meaning its maximum length is strlen("-00000000.00000000") */
  buf_len = strlen(device->cmds->volt_range) + 1 + strlen("-00000000.00000000") + 1 + 1 + 16;
  buf = (char *)malloc(buf_len);
  if (!buf) {
    return -1;
  }
  
  snprintf(buf, buf_len - 1, "%s %8.8f\n", device->cmds->volt_range, value);
  
  /* set the range */
  if(dmm_set_range_finish( handle , &buf) != 0) {
  //fprintf(stderr,"Range_finish failed\n");
  return -1;
  }

  free(buf);
  
  return 0;
}

int dmm_set_res_range (struct dmm_handle *handle, double value) {
  char *buf; /* buffer for the SCPI command */
  int buf_len;
  struct device *device;
  
  device=NULL;
 
  if(dmm_set_range_init( handle , &device ) != 0) {
     // fprintf(stderr,"Range init failed\n");
      return -1;
  }
  
  /* the buffer needs to contain:
     "<device's RANGE command> <range>\n\0<16 bytes of fudge>"
     where <range> is a float of the form %8.8f, meaning its maximum length is strlen("-00000000.00000000") */
  buf_len = strlen(device->cmds->res_range) + 1 + strlen("-00000000.00000000") + 1 + 1 + 16;
  buf = (char *)malloc(buf_len);
  if (!buf) {
    return -1;
  }
  
  snprintf(buf, buf_len - 1, "%s %8.8f\n", device->cmds->res_range, value);
  
  /* set the range */
  if(dmm_set_range_finish( handle , &buf) != 0) {
  //fprintf(stderr,"Range_finish failed\n");
  return -1;
  }

  free(buf);
  
  return 0;
}

#define DMM_READ_BUFFER_SIZE 64

/*
  dmm_read
*/
int dmm_read (struct dmm_handle *handle, double *value) {
  int i;
  char *buf;
  struct device *device;
  int err;
  
  device = &g_devices[handle->device];
  /*sleep(2);*/
  
  /* write out the list of commands required to do a read */
  for (i=0; NULL != device->cmds->read[i]; i++) {
    if (SCPI_ERROR_SUCCESS != scpi_write(handle->scpi, device->cmds->read[i])) {
      return -1;
    }
  }
  
  /* allocate the buffer for reading */
  buf = (char *)malloc(DMM_READ_BUFFER_SIZE + 1);
  
  if (SCPI_ERROR_SUCCESS != (err = scpi_read(handle->scpi, buf, DMM_READ_BUFFER_SIZE, device->cmds->timeout))) {
    //printf("failed to read value, err=%i\n", err);
    free(buf);
    return -1;
  }
  
  *value = strtod(buf, NULL);
  
  free(buf);
  
  return 0;
}

/*
  dmm_get_device_info
*/
int dmm_get_device_info (struct dmm_handle *handle, struct dmm_device_info *devinfo) {
  char *buf;
  int err;
  struct device *device;
  
  device = &g_devices[handle->device];
  
  /* insert default values */
  strcpy(devinfo->manufacturer, "");
  strcpy(devinfo->model, "");
  strcpy(devinfo->serial, "");
  strcpy(devinfo->version, "");
  
  if (NULL == (buf = (char *)malloc(DMM_MAX_SCPI_STRING_LENGTH + 1))) {
    fprintf(stderr,"get_device_info failed at buf malloc\n");
    return -1;
  }
  
 /* if (SCPI_ERROR_SUCCESS != (err = scpi_write(handle->scpi, "*IDN?\n"))) {
   */
    if (SCPI_ERROR_SUCCESS != (err = scpi_write(handle->scpi, device->cmds->ident))) {
    free(buf);
    fprintf(stderr,"scpi_write error in get_device_info \n");
    return -1;
  }
  fprintf(stderr, "Device command sent: %s\n", device->cmds->ident);
  
  if (SCPI_ERROR_SUCCESS != (err = scpi_read(handle->scpi, buf, DMM_MAX_SCPI_STRING_LENGTH, device->cmds->timeout))) {
    free(buf);
    fprintf(stderr,"scpi_read error in get_device_info\n");
    fprintf(stderr, "Arguments: \n DMM_MAX_SCPI_STRING_LENGTH = %i\n",DMM_MAX_SCPI_STRING_LENGTH);
    fprintf(stderr, "Timeout: %i\n",device->cmds->timeout);
    return -1;
  }
  
  /* sscanf'ing into these buffers on the devinfo struct is fine because they are all
     of length DMM_MAX_SCPI_STRING_LENGTH, which is the length of the source buffer ('buf'). */
  
  sscanf(buf, device->cmds->ident_ptrn_str,
	devinfo->manufacturer, devinfo->model, devinfo->serial,
	devinfo->version);
  
  free(buf);
  return 0;
}

/*
  dmm_get_cal_info
*/
int dmm_get_cal_info (struct dmm_handle *handle, struct dmm_cal_info *cal_info) {
  char *buf;
  int err;
  struct device *device;
  
  device = &g_devices[handle->device];
  
  /* insert default values */
  cal_info->date.year = cal_info->date.month = cal_info->date.day = 0;
  cal_info->due.year = cal_info->due.month = cal_info->due.day = 0;
  
  if (NULL == (buf = (char *)malloc(DMM_MAX_SCPI_STRING_LENGTH))) {
    fprintf(stderr,"Malloc failed in get_cal_info\n");
    return -1;
  }
  
  if (SCPI_ERROR_SUCCESS != (err = scpi_write(handle->scpi, device->cmds->cal_query ))) {
    free(buf);
    fprintf(stderr,"scpi_write failed in get_cal_info\n");
    return -1;
  }
  fprintf(stderr, "Device command sent: %s\n", device->cmds->cal_query);
  
  
  err = scpi_read(handle->scpi, buf, DMM_MAX_SCPI_STRING_LENGTH,device->cmds->timeout);
  if(err) {
    fprintf(stderr, "scpi_read error was %i\n", err);
  }
  if (SCPI_ERROR_SUCCESS != err) {
    fprintf(stderr,"scpi_read failed in get_cal_info\n");
    fprintf(stderr,"scpi_read error in get_device_info\n");
    fprintf(stderr, "Arguments: \n DMM_MAX_SCPI_STRING_LENGTH = %i\n",DMM_MAX_SCPI_STRING_LENGTH);
    fprintf(stderr, "Buffer contents: %s \n", buf);
    fprintf(stderr, "Timeout: %i\n",device->cmds->timeout);
    free(buf);
    return -1;
  }
  
 
  sscanf(buf, "%i,%i,%i;%i,%i,%i\n",
	 &cal_info->date.year,
	 &cal_info->date.month,
	 &cal_info->date.day,
	 &cal_info->due.year,
	 &cal_info->due.month,
	 &cal_info->due.day);
  
  free(buf);
  return 0;
}

/*
  dmm_destroy
*/
int dmm_destroy (struct dmm_handle *handle) {
  scpi_destroy(handle->scpi);
  free(handle);
  
  return 0;
}
