#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>

#include "vxi11.h"
#include "scpilib.h"

struct {
  int vxi_error;
  enum scpi_error scpi_error;
} vxi_error_map[] = {
  {0,  SCPI_ERROR_SUCCESS},
  {1,  SCPI_ERROR_SYNTAX_ERROR},
  {2,  SCPI_ERROR_UNKNOWN_ERROR},
  {3,  SCPI_ERROR_DEVICE_NOT_ACCESSIBLE},
  {4,  SCPI_ERROR_INVALID_LINK_IDENTIFIER},
  {5,  SCPI_ERROR_PARAMETER_ERROR},
  {6,  SCPI_ERROR_CHANNEL_NOT_ESTABLISHED},
  {7,  SCPI_ERROR_UNKNOWN_ERROR},
  {8,  SCPI_ERROR_OPERATION_NOT_SUPPORTED},
  {9,  SCPI_ERROR_OUT_OF_RESOURCES},
  {10, SCPI_ERROR_UNKNOWN_ERROR},
  {11, SCPI_ERROR_DEVICE_LINK_LOCKED_BY_ANOTHER_LINK},
  {12, SCPI_ERROR_NO_LOCK_HELD_BY_THIS_LINK},
  {13, SCPI_ERROR_UNKNOWN_ERROR},
  {14, SCPI_ERROR_UNKNOWN_ERROR},
  {15, SCPI_ERROR_IO_TIMEOUT},
  {16, SCPI_ERROR_UNKNOWN_ERROR},
  {17, SCPI_ERROR_IO_ERROR},
  {18, SCPI_ERROR_UNKNOWN_ERROR},
  {19, SCPI_ERROR_UNKNOWN_ERROR},
  {20, SCPI_ERROR_UNKNOWN_ERROR},
  {21, SCPI_ERROR_INVALID_ADDRESS},
  {22, SCPI_ERROR_UNKNOWN_ERROR},
  {23, SCPI_ERROR_ABORT},
  {24, SCPI_ERROR_UNKNOWN_ERROR},
  {25, SCPI_ERROR_UNKNOWN_ERROR},
  {26, SCPI_ERROR_UNKNOWN_ERROR},
  {27, SCPI_ERROR_UNKNOWN_ERROR},
  {28, SCPI_ERROR_UNKNOWN_ERROR},
  {29, SCPI_ERROR_CHANNEL_ALREADY_ESTABLISHED}
};
#define VXI_MIN_ERROR 0
#define VXI_MAX_ERROR 29

#define IO_TIMEOUT 5000 /*milliseconds */
#define LOCK_TIMEOUT 2000 /* milliseconds */

enum scpi_error vxi_error_to_scpi_error (int vxi_error) {
  if (vxi_error < VXI_MIN_ERROR || vxi_error > VXI_MAX_ERROR) {
    return SCPI_ERROR_UNKNOWN_ERROR;
  } else {
    return vxi_error_map[vxi_error].scpi_error;
  }
}

CLIENT *vxi_clnt_create (const char *host, const char *netid) {
  CLIENT *cl;
  
  /* connect to the device */
  cl = clnt_create(host, DEVICE_CORE, DEVICE_CORE_VERSION, netid);
  if (NULL == cl) {
    return NULL;
  }
  
  return cl;
}

enum scpi_error vxi_create_link (CLIENT *cl, Device_Link *lid, const char *device_name) {
  struct Create_LinkResp response = {0, 0, 0, 0};
  struct Create_LinkParms link_parms = {0, 0, 0, NULL};
  int length;
  
  length = strlen(device_name) + 1;
  
  link_parms.clientId = 0; /* TODO: is this needed? */
  link_parms.lockDevice = 0; /* TODO: allow the user to lock the device if needed? */
  link_parms.lock_timeout = LOCK_TIMEOUT;
  link_parms.device = malloc(length);
  memcpy(link_parms.device, device_name, length);
  
  create_link_1(link_parms, &response, cl);
  if (0 != response.error) {
    return vxi_error_to_scpi_error(response.error);
  }
  
  *lid = response.lid;
  
  return vxi_error_to_scpi_error(response.error);
}

enum scpi_error vxi_destroy_link (CLIENT *cl, Device_Link *lid) {
  Device_Error response = {0};
    
  destroy_link_1(*lid, &response, cl);
  if (0 != response.error) {
    return vxi_error_to_scpi_error(response.error);
  }
  
  return vxi_error_to_scpi_error(response.error);
}

int vxi_write (CLIENT *cl, Device_Link lid, const char *data) {
  struct Device_WriteResp response = {0, 0};
  struct Device_WriteParms write_parms = {0, 0, 0, 0, {0, NULL}};
  int length;
  
  length = strlen(data);
  
  write_parms.lid = lid;
  write_parms.io_timeout = IO_TIMEOUT;
  write_parms.lock_timeout = LOCK_TIMEOUT;
  write_parms.flags = 0;
  write_parms.data.data_len = length;
  write_parms.data.data_val = malloc(length);
  memcpy(write_parms.data.data_val, data, length);

  device_write_1(write_parms, &response, cl);
  free(write_parms.data.data_val);
  
  return vxi_error_to_scpi_error(response.error);
}

int vxi_read (CLIENT *cl, Device_Link lid, char *data, int bytes, int timeout) {
  struct Device_ReadResp  response = {0, 0, {0, NULL}};
  struct Device_ReadParms read_parms = {0, 0, 0, 0, 0, 0};
  
  read_parms.lid = lid;
  read_parms.requestSize = bytes - 1;
  read_parms.io_timeout = timeout;
  read_parms.lock_timeout = LOCK_TIMEOUT;
  read_parms.flags = 0x80;
  read_parms.termChar = 0x0a;
  
  device_read_1(read_parms, &response, cl);
  
  if (response.data.data_len) {
    memcpy(data, response.data.data_val, response.data.data_len);
  }
  data[response.data.data_len] = '\0';
  
  free (response.data.data_val);
  return vxi_error_to_scpi_error(response.error);
}

struct scpi_handle *scpi_create (void) {
  struct scpi_handle *handle;
  
  handle = (struct scpi_handle *)malloc(sizeof(struct scpi_handle));
  
  if (NULL == handle) {
    return NULL;
  }
  
  handle->cl = NULL;
  handle->lid = -1;
  handle->error = SCPI_ERROR_UNINITIALIZED;
  
  return handle;
}

enum scpi_error scpi_connect (struct scpi_handle *handle, const char *host, const char *device) {
  int result;
  /* default to no error */
  handle->error = SCPI_ERROR_SUCCESS;
  
  /* connect to the RPC server */
  handle->cl = vxi_clnt_create(host, "tcp");
  if (NULL == handle->cl) {
    handle->error = SCPI_ERROR_CANT_CONNECT;
    return handle->error;
  }
  
  /* create the device link */
  result = vxi_create_link(handle->cl, &(handle->lid), device);
  if (result) {
    handle->error = result;
    return handle->error;
  }
  
  return handle->error;
}

enum scpi_error scpi_write (struct scpi_handle *handle, const char *data) {
  int result;

  /*printf("scpi_write: %s\n", data);*/
  
  /* default to no error */
  handle->error = SCPI_ERROR_SUCCESS;
  
  result = vxi_write(handle->cl, handle->lid, data);
  handle->error = result;
  
  return handle->error;;
}

enum scpi_error scpi_read (struct scpi_handle *handle, char *data, int length, int timeout) {
  int result;

  /* default to no error */
  handle->error = SCPI_ERROR_SUCCESS;
  
  result = vxi_read(handle->cl, handle->lid, data, length, timeout);
  handle->error = result;

  /*printf("scpi_read: %s\n", data);*/
  
  return handle->error;
}

void scpi_destroy (struct scpi_handle *handle) {
  if (NULL == handle) {
    return;
  }
  
  if (handle) {
    /* destroy the link */
    vxi_destroy_link( handle->cl, &handle->lid );
    if (handle->cl) {
      clnt_destroy(handle->cl);
    }
  }
  free (handle);
}
